use std::fmt::Display;

use glam::{Mat3, Vec3, Vec4};

use crate::math;

//? Should this be normalised or does that risk introducing fpp errors
#[derive(Clone, Copy, Debug, Default)]
pub struct Plane {
    pub normal: Vec3,
    pub offset: f32,
}

impl Plane {
    pub fn new(normal: Vec3, offset: f32) -> Self {
        Self { normal, offset }
    }

    pub fn from_point_normal(point: Vec3, normal: Vec3) -> Self {
        Self {
            normal,
            offset: point.dot(normal),
        }
    }

    pub fn point_in_halfspace(&self, point: Vec3) -> bool {
        self.normal.dot(point) - self.offset <= math::EPSILON
    }

    pub fn into_vec4(&self) -> Vec4 {
        glam::vec4(self.normal.x, self.normal.y, self.normal.z, -self.offset)
    }
}

impl From<Vec4> for Plane {
    fn from(value: Vec4) -> Self {
        Self {
            normal: Vec3::new(value.x, value.y, value.z),
            offset: -value.w,
        }
    }
}

impl PartialEq for Plane {
    fn eq(&self, other: &Self) -> bool {
        (self.normal.x - other.normal.x).abs() < math::EPSILON
            && (self.normal.y - other.normal.y).abs() < math::EPSILON
            && (self.normal.z - other.normal.z).abs() < math::EPSILON
            && (self.offset - other.offset).abs() < math::EPSILON
    }
}

impl Eq for Plane {}

impl Display for Plane {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(
            f,
            "({})x + ({})y +({})z + ({}) = 0",
            self.normal.x, self.normal.y, self.normal.z, -self.offset
        )
    }
}

#[derive(Clone, Copy, Debug)]
pub struct PlaneIntersection {
    pub planes: [Plane; 3],
}

impl PlaneIntersection {
    pub fn from_planes(p1: Plane, p2: Plane, p3: Plane) -> Option<Self> {
        let plane = Self {
            planes: [p1, p2, p3],
        };

        let determinant = plane.get_matrix().determinant();
        // println!("Determinant: {determinant}");
        if determinant.abs() > math::EPSILON {
            Some(plane)
        } else {
            None
        }
    }

    pub fn get_intersection_point(&self) -> Vec3 {
        let base_matrix = self.get_matrix();
        let base_det = base_matrix.determinant();
        let offsets = glam::vec3(
            self.planes[0].offset,
            self.planes[1].offset,
            self.planes[2].offset,
        );

        let mut vs = vec![];
        for idx in 0..3 {
            let mut matrix = base_matrix;
            *(matrix.col_mut(idx)) = offsets;
            vs.push(matrix.determinant() / base_det);
        }

        Vec3::from_slice(&vs)
    }

    pub fn in_halfspace_mat(&self, plane: &Plane) -> bool {
        // "Distance" and side of the point defined by plane against the plane defined
        // by the 3 points of the intersection planes.
        //* Note that the sign is dependant on the winding order of the intersection planes
        let m1 = glam::mat4(
            self.planes[0].into_vec4(),
            self.planes[1].into_vec4(),
            self.planes[2].into_vec4(),
            plane.into_vec4(),
        );
        let d1 = m1.determinant();

        // We can resolve the winding order problem by multiplying by the normal matrix determinant
        let m2 = glam::mat3(
            self.planes[0].normal,
            self.planes[1].normal,
            self.planes[2].normal,
        );
        let d2 = m2.determinant();
        let dist = d1 * d2;

        //* Note: dist is only euclidean accurate if all the normals are normalised :)
        dist < math::EPSILON
    }

    fn get_matrix(&self) -> Mat3 {
        glam::mat3(
            self.planes[0].normal,
            self.planes[1].normal,
            self.planes[2].normal,
        )
        .transpose()
    }
}

#[cfg(test)]
mod plane_tests {
    use super::*;

    #[test]
    fn from_point_normal() {
        let point = glam::vec3(1.0, 1.0, 0.0);
        let normal = glam::vec3(1.0, 1.0, 1.0);

        let expected = Plane::new(normal, 2.0);

        assert_eq!(Plane::from_point_normal(point, normal), expected);
    }

    #[test]
    fn point_in_halfspace() {
        let intersection = PlaneIntersection::from_planes(
            Plane::from_point_normal(Vec3::X, Vec3::X),
            Plane::from_point_normal(Vec3::Y, Vec3::Y),
            Plane::from_point_normal(Vec3::Z, Vec3::Z),
        )
        .unwrap();
        let point = intersection.get_intersection_point();

        let p1 = Plane::from_point_normal(Vec3::NEG_ONE, Vec3::NEG_ONE);
        let p2 = Plane::from_point_normal(Vec3::ZERO, Vec3::ONE);
        assert!(p1.point_in_halfspace(point));
        assert!(!p2.point_in_halfspace(point));
    }
}

#[cfg(test)]
mod plane_intersection_tests {
    use super::*;

    #[test]
    fn from_planes() {
        let planes = vec![
            Plane::from_point_normal(Vec3::X, Vec3::X),
            Plane::from_point_normal(Vec3::Y, Vec3::Y),
            Plane::from_point_normal(Vec3::Z, Vec3::Z),
            Plane::from_point_normal(Vec3::NEG_X, Vec3::NEG_X),
        ];

        let intersection1 = PlaneIntersection::from_planes(planes[0], planes[1], planes[2]);
        let intersection2 = PlaneIntersection::from_planes(planes[0], planes[1], planes[3]);

        assert!(intersection1.is_some());
        assert!(intersection2.is_none());
    }

    #[test]
    fn get_intersection_point() {
        let planes = vec![
            Plane::from_point_normal(Vec3::X, Vec3::X),
            Plane::from_point_normal(Vec3::Y, Vec3::Y),
            Plane::from_point_normal(Vec3::Z, Vec3::Z),
            Plane::from_point_normal(glam::vec3(1.0, 1.0, 0.0), Vec3::ONE),
        ];

        let intersection1 = PlaneIntersection::from_planes(planes[0], planes[1], planes[2]);
        let intersection2 = PlaneIntersection::from_planes(planes[0], planes[1], planes[3]);
        assert_eq!(intersection1.unwrap().get_intersection_point(), Vec3::ONE);
        assert_eq!(
            intersection2.unwrap().get_intersection_point(),
            glam::vec3(1.0, 1.0, 0.0)
        );
    }

    #[test]
    fn in_half_space_mat() {
        let intersection = PlaneIntersection::from_planes(
            Plane::from_point_normal(Vec3::X, Vec3::X),
            Plane::from_point_normal(Vec3::Y, Vec3::Y),
            Plane::from_point_normal(Vec3::Z, Vec3::Z),
        )
        .unwrap();

        let p1 = Plane::from_point_normal(Vec3::NEG_ONE, Vec3::NEG_ONE);
        let p2 = Plane::from_point_normal(Vec3::ZERO, Vec3::ONE);
        assert!(intersection.in_halfspace_mat(&p1));
        assert!(!intersection.in_halfspace_mat(&p2));
    }
}
